/** 
 @file events.js
 
 @brief This file contain implementation of the event handling from the upnp ui.

 @author: mobile2DS

 @date: 15.12.2013
*/
var lWarn = {logln:function (o){}}; 
var lInfo = {logln:function (o){}};

var LOGFILE_NAME = "sonos-controller.log";
lWarn = new Logger(LOGFILE_NAME);
lInfo = new Logger(LOGFILE_NAME);

/*
 returns id of players master if current player is slave, if current player is master result is empty
*/
function getPlayerGroupCoordinator(aGroups, sPlayerIP, sDeviceID) {
  sCoordinator = "";
  if (aGroups != null) {
    for (var iIndex=0;iIndex<aGroups.length;iIndex++) {
      var sGroup = aGroups[iIndex].data; 
      //lInfo.logln(sGroup);
          if (sGroup.indexOf(sPlayerIP) > -1) {
            var sMaster = sGroup.match(/Coordinator=\"(.*?)\"/); 
            //lInfo.logln("sMaster="+sMaster+"/"+sDeviceID);
            if (sMaster != null) {
               if (sMaster[1] != sDeviceID && sMaster[1].indexOf('RINCON') >-1) {
                sCoordinator = sMaster[1];
               }
            } 
            break;
          }                     
    } 
  }
  return sCoordinator;
}  

function playRadio(sPlayerIP, sUrl, sTitle, curl_options) {
  upnpSetRadio(sPlayerIP, sUrl, sTitle, curl_options);
  upnpPlay(sPlayerIP, curl_options);
}

function playQueue(sPlayerIP, sDeviceID, execAction, curl_options) {
  var arrayQueue = execAction.getChild('active-queue').getChildren();
  if (arrayQueue!=null) {
    lInfo.logln("Queue"+arrayQueue.length);
    upnpClearQueue(sPlayerIP, curl_options);
    for (var yIndex=0;yIndex<arrayQueue.length;yIndex++) {
      upnpLoadPlaylist(sPlayerIP, arrayQueue[yIndex].getChild('playID').getValue(), curl_options);
      if (yIndex == 0) { // start playing first song
        upnpSetContentSrc(sPlayerIP, "x-rincon-queue:"+sDeviceID+"#0", curl_options);
        //upnpSetContentSrc(sPlayerIP, arrayQueue[yIndex].getChild('playID').getValue(), curl_options);
        upnpPlay(sPlayerIP, curl_options);
      }
      lInfo.logln("Adding "+arrayQueue[yIndex].getChild('playID').getValue()); 
    }
  }
}

function playPlaylist(sPlayerIP, sDeviceID, sFrom, execAction, curl_options, sMediaGUID) {
  upnpClearQueue(sPlayerIP, curl_options);
  var sLoadValue = execAction.getChild(sFrom).getValue();
  if (sLoadValue != "") {
    upnpLoadPlaylist(sPlayerIP, sLoadValue, curl_options);
    
    // activating queue, in not active
    upnpSetContentSrc(sPlayerIP, "x-rincon-queue:"+sDeviceID+"#0", curl_options);
    upnpSeek(sPlayerIP, true, 1, curl_options);
    upnpPlay(sPlayerIP, curl_options);
  }
  else
    if (sMediaGUID != "" && execAction.getChild(sMediaGUID) != null) {  // media file withouth url
      var sGuidID = execAction.getChild(sMediaGUID).getValue();
      if (sGuidID != "") {
        cloudService.fetchMediaFile(unescape(sGuidID), function (oData) {
          if (oData.ReturnMessage == "OK") {
            var sUrl = unescape(oData.Response.UrlPath); 
            storeSubParameter('sonos-activities', execAction.getChild("id").getValue()+'/nowSingleUrl', sUrl);
            
            upnpLoadPlaylist(sPlayerIP, sUrl, curl_options);
            
            // activating queue, in not active
            upnpSetContentSrc(sPlayerIP, "x-rincon-queue:"+sDeviceID+"#0", curl_options);
            upnpSeek(sPlayerIP, true, 1, curl_options);
            upnpPlay(sPlayerIP, curl_options);
            
            // nastavimo še dolžino komada
            objCurrentTrack = upnpGetCurrentTrack(sPlayerIP, curl_options);
            if (objCurrentTrack != null) {
              var nLength = secondToInt(unescape(objCurrentTrack.length));
              if (!isNaN(nLength))
                storeSubParameter('sonos-activities', execAction.getChild('id').getValue()+'/nowSingleLength', nLength);
            }
          }
        });
      }
    }
}

function setPlayerVolume(sPlayerIP, execAction, curl_options, sGroupedSlavesIP, bSkipStop) {
   var bActionVolume = JSON.parse(execAction.getChild('volumeSet').getValue());
   if (bActionVolume != null  && JSON.parse(bActionVolume)) {
     var nNewVolume =  parseInt(execAction.getChild('volume').getValue());
     if (nNewVolume != null)
       if (!bSkipStop)
         upnpStop(sPlayerIP, curl_options);
       
       lInfo.logln("Exec command set volume "+nNewVolume);
       upnpSetVolume(sPlayerIP, nNewVolume, curl_options);
       
       // set volume also for all slave players in group
       var aSlaves = null;
       if (sGroupedSlavesIP.length > 0) {
           aSlaves = sGroupedSlavesIP.split(","); 
       }
       
       if (aSlaves != null && aSlaves.length > 0) {
         for (var iIndex=0;iIndex<aSlaves.length;iIndex++) {
           var sSlaveIP = aSlaves[iIndex].trim();
           lInfo.logln("Exec command set slave volume for "+sSlaveIP);
           upnpSetVolume(sSlaveIP, nNewVolume, curl_options);
         }
       }
   }
}

function executeComandOnSonos(sPlayerIP, sDeviceID, execAction, curl_options, sGroupedSlavesIP, sRootPath) {
  var nActionType = parseInt(execAction.getChild('actionType').getValue());
  lInfo.logln("nActionType="+nActionType);
  
  var bSetVolume = true;
  if(nActionType == SONOS_ACTION_COMMAND) {
    var nCommandID = parseInt(execAction.getChild('command').getValue());                 
    if (nCommandID == COMMAND_MUTE || nCommandID == COMMAND_UNMUTE)
      bSetVolume = false;
  }
  
  if (bSetVolume) {
    setPlayerVolume(sPlayerIP, execAction, curl_options, sGroupedSlavesIP, false);
  }
  
  switch (nActionType) {
    case SONOS_ACTION_NOW_PLAYING: // Now playing
      lInfo.logln("Now playing");
      if (execAction.getChild('nowAction').getValue() == 0) { // Single track/Radio
        lInfo.logln("Set "+execAction.getChild('playListURL').getValue());   
        if (JSON.parse(execAction.getChild('nowSingleIsRadio').getValue())) { // radio
          lInfo.logln("Play radio "+execAction.getChild('nowSingleTitle').getValue());   
          playRadio(sPlayerIP, execAction.getChild('nowSingleUrl').getValue(), execAction.getChild('nowSingleTitle').getValue(), curl_options);
        }
        else { // mp3  
           playPlaylist(sPlayerIP, sDeviceID, 'nowSingleUrl', execAction, curl_options, "");
        }
      }
      else { // Selected queue
        playQueue(sPlayerIP, sDeviceID, execAction, curl_options);
      }
    break;

    case SONOS_ACTION_PLAYLIST: // Playlist
      lInfo.logln("Playlist");                      
      playPlaylist(sPlayerIP, sDeviceID, 'playListURL', execAction, curl_options, "");
    break;

    case SONOS_ACTION_FAVORITES: // Sonos favorites
      lInfo.logln("Favorites");          
      // radio favorite
      if (execAction.getChild('favoritesTitle').getValue().indexOf("TuneIn Station") > -1) {
         lInfo.logln("Favorites radio");                                                           
         playRadio(sPlayerIP, execAction.getChild('favoritesURL').getValue(), execAction.getChild('favoritesTitle').getValue(), curl_options);
      }
      else {
        lInfo.logln("Favorites other");          
        // other favorites: replace queue, add to queue, play
        playPlaylist(sPlayerIP, sDeviceID, 'favoritesURL', execAction, curl_options, "");
        
        // if single mp3 is selected in favorites, we must fetch file lenght (only if not defined yet). 
        // It's not possible to get this info when saving activity in dialog
        var sLength = execAction.getChild('nowSingleLength').getValue();
        if (sLength == "" || parseInt("" + sLength,10) < 1) {
          objCurrentTrack = upnpGetCurrentTrack(sPlayerIP, curl_options);
          if (objCurrentTrack != null) {
            var nLength = secondToInt(unescape(objCurrentTrack.length));
            if (!isNaN(nLength))
              storeSubParameter(sRootPath, execAction.getChild('id').getValue()+'/nowSingleLength', nLength);
          }
        } 
      }
    break;
    
    case SONOS_ACTION_LIBRARY:
    case SONOS_ACTION_TTS:
      lInfo.logln("MediaLibrary/Text2Speech");
      
      var sGUID = "mediaFileGUID";
      if (nActionType == SONOS_ACTION_TTS)
        sGUID = "t2Sguid";
      
      playPlaylist(sPlayerIP, sDeviceID, 'nowSingleUrl', execAction, curl_options, sGUID);

      var sLength = execAction.getChild('nowSingleLength').getValue();
      if (sLength == "" || parseInt("" + sLength,10) < 1) {
        objCurrentTrack = upnpGetCurrentTrack(sPlayerIP, curl_options);
        if (objCurrentTrack != null) {
          var nLength = secondToInt(unescape(objCurrentTrack.length));
          if (!isNaN(nLength))
            storeSubParameter(sRootPath, execAction.getChild('id').getValue()+'/nowSingleLength', nLength);
        }
      } 

    break;

    case SONOS_ACTION_COMMAND: // Command
      lInfo.logln("command="+nCommandID);      
      switch(nCommandID) {
        case COMMAND_STOP: //Stop 
          upnpStop(sPlayerIP, curl_options);
        break;

        case COMMAND_PLAY: // Play
          upnpPlay(sPlayerIP, curl_options);
        break;
                                  
        case COMMAND_PAUSE: // Pause
          upnpPause(sPlayerIP, curl_options);
        break;

        case COMMAND_MUTE: // Mute
          upnpSetMute(sPlayerIP, true, curl_options);
          bSetVolume = false;
        break;

        case COMMAND_UNMUTE: // Unmute
          upnpSetMute(sPlayerIP, false, curl_options);
        break;
      }
    break;
  }
}

function getPlayersToExecUndo(oActivity) {
  var arrayPlayers = new Array();
  if (oActivity != null) {
    if (JSON.parse(oActivity.getChild('allRooms').getValue())) {// all players
       var baseNode = Property.getNode('sonos-players');
       if (baseNode != null) {
         var oArrayNode = baseNode.getChildren();
         for ( var iIndex = 0; iIndex < oArrayNode.length; iIndex++) {
           if (!JSON.parse(oArrayNode[iIndex].getChild('disabled').getValue())) 
             arrayPlayers.push(new Object({id:oArrayNode[iIndex].getChild('id').getValue(),
                                           device:oArrayNode[iIndex].getChild('deviceID').getValue(),
                                           ip:oArrayNode[iIndex].getChild('ipAddress').getValue()
                                          }));
         }
       }
    }
    else {
      // only selected players/rooms
      var baseNode = oActivity.getChild('pickedRooms');
      if (baseNode != null && baseNode.getChild('pickedRooms') != null) {
        var sRooms = baseNode.getChild('pickedRooms').getValue();
        if (sRooms != null) {
          var rooms = unescape(sRooms).split(";");
          var oArrayNode = Property.getNode('sonos-players').getChildren();
          for ( var iIndex = 0; iIndex < oArrayNode.length; iIndex++) {
            if (!JSON.parse(oArrayNode[iIndex].getChild('disabled').getValue())) {
              if (rooms.indexOf(oArrayNode[iIndex].getChild('roomID').getValue())>-1)
                arrayPlayers.push(new Object({id:oArrayNode[iIndex].getChild('id').getValue(),
                                            device:oArrayNode[iIndex].getChild('deviceID').getValue(),
                                            ip:oArrayNode[iIndex].getChild('ipAddress').getValue()}));
            }
          }
        }
      }
    }
  }
  return arrayPlayers;
}

function resetDimming() {
  var dimmValue = Property.getNode("dimming");
  if (dimmValue != null) {
    var oArrayNode = dimmValue.getChildren();
    for ( var iIndex = 0; iIndex < oArrayNode.length; iIndex++) {
      var oChildNode = oArrayNode[iIndex];
      dimmValue.removeChild(oChildNode);
    }
  }
}

var curl_options = new Object({'debug' : 1, 'connectTimeout' : 200}); 
lInfo.logln(JSON.stringify(raisedEvent)); 

if (raisedEvent.name === 'sonos_execute' || raisedEvent.name === 'sonos_execute_pause') {
 var start = new Date().getTime();
  try {  
    if (raisedEvent.parameter != null) {
      var nGroupID = -1;
      if (raisedEvent.parameter.groupID != null)
         nGroupID = parseInt(""+raisedEvent.parameter.groupID,10);
      lInfo.logln("nGroupID  " +  nGroupID);
      lInfo.logln("path  " +  raisedEvent.parameter.path);
      
      // reseting dimming tmp values
      resetDimming();
        
      var execAction = Property.getNode(raisedEvent.parameter.path);
      if (execAction != null) {
        if (nGroupID == 4 || raisedEvent.name === 'sonos_execute_pause') 
        { // check audio actions
          var nZoneID = parseInt(execAction.getChild('roomID').getValue());
          var baseNode = Property.getNode('sonos-players');
          var arrayPlayers = new Array();
          if (baseNode != null) {
            var arrayPlayers = baseNode.getChildren();
            var sPlayerIP = "";
            var sDeviceID = "";
            var sGroupedSlavesIP = "";
            if (arrayPlayers!=null) {
              for (var iIndex=0;iIndex<arrayPlayers.length;iIndex++) {
                  if (arrayPlayers[iIndex].getChild('roomID').getValue() == nZoneID) {
                    if (!JSON.parse(arrayPlayers[iIndex].getChild('disabled').getValue())) {
                      sPlayerIP = arrayPlayers[iIndex].getChild('ipAddress').getValue();
                      sDeviceID = arrayPlayers[iIndex].getChild('deviceID').getValue();
                      if (arrayPlayers[iIndex].getChild('playersListIPSlaves') != null) 
                        sGroupedSlavesIP = arrayPlayers[iIndex].getChild('playersListIPSlaves').getValue();
                      break;
                    }
                  }
              }
            }
            lInfo.logln("zone["+nZoneID+"], assigned sPlayerIP="+sPlayerIP+"/id="+sDeviceID);
            // test if player moved to other group, then assign to master room
            arraySonosGroups = upnpGetSonosGroups(sPlayerIP, curl_options);
            var sCoordinator = getPlayerGroupCoordinator(arraySonosGroups, sPlayerIP, sDeviceID);
            lInfo.logln("sCoordinator="+sCoordinator);
            //var sCoordinator = upnpGetGroupCoordinator(sPlayerIP, sDeviceID, curl_options);
            if (sCoordinator.length > 0) { 
               // room player is part of group and not coordinator, we set sPlayerIP to coordinator of this group 
              for (var iIndex=0;iIndex<arrayPlayers.length;iIndex++) {
                if (arrayPlayers[iIndex].getChild('deviceID').getValue() == sCoordinator) {
                  sPlayerIP = arrayPlayers[iIndex].getChild('ipAddress').getValue();
                  sDeviceID = sCoordinator;
                  break;
                }
              }
            }                                                            
            lInfo.logln("action execution sPlayerIP="+sPlayerIP);
            if (sPlayerIP != "") {
              executeComandOnSonos(sPlayerIP, sDeviceID, execAction, curl_options, sGroupedSlavesIP, 'sonos-presets');
              upnpSetPlayMode(sPlayerIP, "NORMAL", curl_options);
            }
          }
        }   
        else {
          var sendAction = null;
          var listPlayers = new Array();
          
          var sGroupingActive = "";
          var baseNode = Property.getNode('sonos-grouping');
          if (baseNode != null) 
            sGroupingActive = baseNode.getChild('data').getValue();

          if (sGroupingActive == "")  {
            lInfo.logln("Scene recieved="+raisedEvent.parameter.sceneID);            
            
            // default as costum event handling
            var nSceneID = 0; 
            var nEventType = EVENT_TYPE_CUSTOM;

            if (typeof raisedEvent.parameter.sceneID !== 'undefined') {
              nSceneID = parseInt(raisedEvent.parameter.sceneID);
              nEventType = EVENT_TYPE_CALL_SCENE;
            }
            else {
              if (typeof raisedEvent.parameter.statename !== 'undefined') {
                nSceneID = raisedEvent.parameter.statename;
                if (typeof raisedEvent.parameter.state !== 'undefined') 
                  nSceneID += raisedEvent.parameter.state;
                
                nEventType = EVENT_TYPE_CALL_SCENE;
                // undefined sceneID's  
                switch(nSceneID){
                  case 'fireactive':
                    nSceneID = 76;
                  break;
                  case 'rainactive':
                    nSceneID = 88;
                  break;
                  case 'raininactive':
                    nSceneID = 89;
                  break;
                  case 'hailactive':
                    nSceneID = 90;
                  break;
                  case 'hailinactive':
                    nSceneID = 91;
                  break;
                  case 'windactive':
                    nSceneID = 86;
                  break;
                  case 'windinactive':
                    nSceneID = 87;
                  break;
                  default:
                    nEventType = EVENT_TYPE_CUSTOM;
                    nSceneID = 0;
                  break;
                }
              }
            }
              
            lInfo.logln("Scene final= " + nSceneID + ", type " + nEventType);
            var bSpecialScenesOnlyRestore = false; // for sceneID > 64

            sendAction = new Object();
            sendAction.volumeSet    = JSON.parse(execAction.getChild('volumeSet').getValue());
            sendAction.volume       = execAction.getChild('volume').getValue();
            sendAction.actionType   = execAction.getChild('actionType').getValue();
            sendAction.command          = execAction.getChild('command').getValue();
            sendAction.restartAfter     = JSON.parse(execAction.getChild('restartAfter').getValue());
            if ((!sendAction.restartAfter) && (nSceneID > 64) 
                && (nEventType == EVENT_TYPE_CALL_SCENE)) {
              sendAction.restartAfter = true;
              bSpecialScenesOnlyRestore = true;
            }
              
            sendAction.repeatLimited    = JSON.parse(execAction.getChild('setRepeat').getValue());
            sendAction.createSonosGroup = false;
            if (execAction.getChild('createSonosGroup') != null)
              sendAction.createSonosGroup = JSON.parse(execAction.getChild('createSonosGroup').getValue());
            
            sendAction.repeatTime       = 0;
            if (sendAction.repeatLimited)
              sendAction.repeatTime     = parseInt(""+execAction.getChild('repeatTime').getValue(),10)*60; // repeat for maximum of x minutes, converted to seconds
    
            sendAction.playInAllRooms   = JSON.parse(execAction.getChild('allRooms').getValue());
            sendAction.filteredRooms    = new Array();
            if (!sendAction.playInAllRooms) 
              sendAction.filteredRooms = unescape(execAction.getChild('pickedRooms').getValue()).split(';');
            
            if (nSceneID == -1) {// test button called from UI
              sendAction.restartAfter = true; // allways restart
              sendAction.repeatTime = TEST_PLAY_TIME; // test time before restore
            }
            
            if (sendAction != null) {
              // 2. list of players to send command 
              var baseNode = Property.getNode('sonos-players');
              if (baseNode != null) {
                var arraySlavePlayers = new Array();
                var arrayPlayers = baseNode.getChildren();
                if (arrayPlayers!=null) {
                  var arraySonosGroups = null;
                  var arrayFilteredPlayers = new Array();
                  for (var iIndex=0;iIndex<arrayPlayers.length;iIndex++) {
                    var playerData = arrayPlayers[iIndex];
                    if (!JSON.parse(playerData.getChild('disabled').getValue())) {
                      var playerRoomID = playerData.getChild('roomID').getValue();
                      var sPlayerIP = playerData.getChild('ipAddress').getValue();
                      var playerID = playerData.getChild('id').getValue();
                      var sDeviceID = playerData.getChild('deviceID').getValue();
                      var sIPSlaves = playerData.getChild('playersListIPSlaves').getValue();

                      if ( sendAction.playInAllRooms ||
                          (!sendAction.playInAllRooms && sendAction.filteredRooms.indexOf(playerRoomID.toString())>-1) // skipping rooms that are not included in activity
                         ) {
                           if (arraySonosGroups == null) { // reading all sonoses groups only for the first time  
                             arraySonosGroups = upnpGetSonosGroups(sPlayerIP, curl_options);
                           }
                           var sCoordinator = getPlayerGroupCoordinator(arraySonosGroups, sPlayerIP, sDeviceID);

                           if (sCoordinator.length > 0) { // current player is in group and not master, so we skip him
                             /*
                             lInfo.logln("Player["+sPlayerIP+"] is coordinator of "+sCoordinator);                   
                             continue; // other player is group master
                             */
                              // room player is part of group and not coordinator, we set sPlayerIP to coordinator of this group 
                             for (var xIndex=0;xIndex<arrayPlayers.length;xIndex++) {
                               if (arrayPlayers[xIndex].getChild('deviceID').getValue() == sCoordinator) {
                                 sPlayerIP = arrayPlayers[xIndex].getChild('ipAddress').getValue();
                                 sDeviceID = sCoordinator;
                                 break;
                               }
                             }
                           }
                           
                           var handleResult = handleNewActivity(playerID, 0
                                                                , nSceneID, nEventType
                                                                , COMMAND_PLAY
                                                                , sendAction.restartAfter);
                           if (!handleResult.storeActivity) {
                             var bStateOk = true;
                             if (sendAction.restartAfter && handleResult.storeCurrDevData) { 
                               restartData = new Object();
                               restartData.id = playerData.getChild('id').getValue();
                               restartData.isRadio = true;
                               restartData.volumeDef = "";
                               restartData.playMode = "";
                               restartData.trackListPos = "";
                               restartData.trackPlayPos = "";
                               restartData.sceneID = nSceneID;                           
                               restartData.playerIP = sPlayerIP;
                               restartData.volumeReset = true;
                               restartData.activeAction = "STOPPED";
                               if (!bSpecialScenesOnlyRestore)
                                 restartData.activeAction = upnpGetTransportInfo(sPlayerIP, curl_options);
                               restartData.activeQueueID = upnpSaveQueue(sPlayerIP, playerID, curl_options);
                               
                               var sMedia = upnpGetMediaInfo(sPlayerIP, curl_options);
                               matches = sMedia.match(/<CurrentURI>(.*?)<\/CurrentURI>/); 
                               var nrTracks = 0;
                               if (matches != null) {
                                 restartData.playUrl = matches[1];
                                 if (matches[1].indexOf("x-rincon-queue")>-1) { // active queue
                                   restartData.isRadio = false;
                                   matches = sMedia.match(/<NrTracks>(.*?)<\/NrTracks>/);
                                   if (matches != null) 
                                     nrTracks = parseInt("" + matches[1],10);
                                   
                                   lInfo.logln("nrTracks="+nrTracks);                         
                                   if (nrTracks > 0) { // remembering current track nr., music pos.
                                     var sPosInfo = upnpGetPositionInfo(sPlayerIP, curl_options);                            

                                     restartData.trackListPos = 1;
                                     matches = sPosInfo.match(/<Track>(.*?)<\/Track>/);
                                     if (matches != null) 
                                       restartData.trackListPos = parseInt("" + matches[1],10);
                                     lInfo.logln("trackListPos="+restartData.trackListPos);                         
                                       
                                     restartData.trackPlayPos = "0:00:00";
                                     matches = sPosInfo.match(/<RelTime>(.*?)<\/RelTime>/);
                                     if (matches != null) 
                                       restartData.trackPlayPos = matches[1];
                                     lInfo.logln("trackPlayPos="+restartData.trackPlayPos);                         
                                   }
                                 }     
                               }   
                                             
                               restartData.volumeDef = upnpGetVolume(sPlayerIP, curl_options);
                               // store slaves volume values
                               var slavesVolumes = "";
                               if (sIPSlaves.length > 0) {
                                 aSlaves = sIPSlaves.split(","); 
                               
                                 if (aSlaves != null && aSlaves.length > 0) {
                                   for (var xIndex=0;xIndex<aSlaves.length;xIndex++) {
                                     var sSlaveIP = aSlaves[xIndex].trim();
                                     var nTmpVolume = upnpGetVolume(sSlaveIP, curl_options);
                                     if (slavesVolumes.length>0)
                                       slavesVolumes = slavesVolumes + ",";
                                     if (nTmpVolume != null)
                                       slavesVolumes = slavesVolumes + nTmpVolume
                                     else
                                       slavesVolumes = slavesVolumes + "-1"; // error by reading value
                                   }
                                 } 
                               }
                               restartData.volumeSlaves = slavesVolumes;
                               restartData.playMode = upnpGetPlayMode(sPlayerIP, curl_options);
                               
                               bStateOk = storeUpnpDeviceStatus(restartData);
                             }
                             
                             if (handleResult.executeUpnpCmd && bStateOk)                                                                             
                               arrayFilteredPlayers.push(new Object({"id":playerID, 
                                                                     "ip":sPlayerIP, 
                                                                     "device":sDeviceID, 
                                                                     "slaves":playerData.getChild('playersListSlaves').getValue(),
                                                                     "IPSlaves":sIPSlaves,
                                                                    }
                                                                   )
                                                        );
                           }
                           else {
                             lInfo.logln("Store only activity");
                             storeActivityForLater(playerID, 0, execAction.getChild('id').getValue());
                           }                       
                         }
                    }
                    else {
                      lInfo.logln("Player("+playerData.getChild('sonosRoom').getValue()+") skipped while not selected as active");
                    }
                  }

                  var sMasterGroupDevice = "";
                  for (var iIndex=0;iIndex<arrayFilteredPlayers.length;iIndex++) {
                    var playerID = arrayFilteredPlayers[iIndex].id;
                    var sPlayerIP = arrayFilteredPlayers[iIndex].ip;
                    var sDeviceID = arrayFilteredPlayers[iIndex].device;
                    var sGroupedSlavesIP = arrayFilteredPlayers[iIndex].IPSlaves;

                    if (iIndex == 0) 
                      sMasterGroupDevice = sDeviceID;
                    
                    lInfo.logln("Execute on player with IP "+sPlayerIP);
                    if (sendAction.createSonosGroup && iIndex > 0) {// 
                      // connect sonos to temporary group
                      lInfo.logln("Assign player["+sPlayerIP+"] to group "+sMasterGroupDevice);
                      // add player to group
                      upnpSetContentSrc(sPlayerIP, "x-rincon:"+sMasterGroupDevice, curl_options);
                      
                      //check if has slaves, add slaves to group
                      addSlavePlayersToGroup(arrayFilteredPlayers[iIndex].slaves, arraySonosGroups, sMasterGroupDevice, curl_options);
                      // must set volume for each slave parent separately
                      setPlayerVolume(sPlayerIP, execAction, curl_options, "", true);
                    }
                    else
                    {
                      var isSceneWithUndo = (predefinedUnduEvents.indexOf(nSceneID) > -1)
                                                                                    
                      //lInfo.logln("sendAction: " + JSON.stringify(sendAction));
                      executeComandOnSonos(sPlayerIP, sDeviceID, execAction, curl_options, sGroupedSlavesIP, 'sonos-activities');
                      var sPlayerMode = upnpGetPlayMode(sPlayerIP, curl_options);
                      
                      if (sendAction.repeatLimited) { // || isSceneWithUndo
                        upnpSetPlayMode(sPlayerIP, "REPEAT_ALL", curl_options);
                      }
                      else
                        if (sPlayerMode != 'NORMAL')
                          upnpSetPlayMode(sPlayerIP, "NORMAL", curl_options);
                      
                      // we set this value after execution command to sonos, in case of empty legth of mp3, we set it there
                      sendAction.nowSingleLength = parseInt(""+execAction.getChild('nowSingleLength').getValue(),10);
                     
                      if (sendAction.repeatLimited || (sendAction.createSonosGroup && !isSceneWithUndo)  ||
                          (sendAction.restartAfter && !isSceneWithUndo)
                         ) { // set timer if time limited execution, or not undo event with restart data
                        var execLength = sendAction.repeatTime;
                        
                        if (nSceneID != -1) { // if not testing button
                          if (sendAction.nowSingleLength > 0 && !sendAction.repeatLimited ) {  // defined exec time by mp3 file time
                            lInfo.logln("Set new Length="+sendAction.nowSingleLength);
                            execLength = sendAction.nowSingleLength;
                          }
                        }
                         
                        // in case of length problem we set predefined default exec time   
                        if (execLength == 0)
                          execLength = DEF_REPEAT_TIME;

                        if (!sendAction.createSonosGroup) { // no player grouping, create event for each speaker
                          var timerData = new Object();
                          timerData.sceneID = nSceneID;
                          timerData.id = playerID;
                          timerData.eventType = nEventType;
                          timerData.playerIP = sPlayerIP;
                          timerData.deviceID = arrayFilteredPlayers[iIndex].device;
                          timerData.onlyStop = (sendAction.repeatLimited && !sendAction.restartAfter);
                          
                          lInfo.logln("New timer will execeute after ="+execLength); 
                          
                          (new Event('sonos_execute_timer',{path:JSON.stringify(timerData), time : '+'+execLength })).raise();
                        }
                        else {
                          // players grouping create single event
                          var timerData = new Object();
                          timerData.sceneID = nSceneID;
                          timerData.eventType = nEventType;
                          timerData.playerMasters = arrayFilteredPlayers;
                          timerData.onlyStop = (sendAction.repeatLimited && !sendAction.restartAfter);
                          
                          lInfo.logln("New group timer will execeute after ="+execLength); 
                          
                          (new Event('sonos_execute_timer_group',{path:JSON.stringify(timerData), time : '+'+execLength })).raise();
                        }
                      }
                    } 
                  }
                }
              }    
              
              
              
              if (sendAction.createSonosGroup && arrayFilteredPlayers.length>0) {
                var sGroupedData = escape(JSON.stringify(arrayFilteredPlayers));  
                storeSubParameter('sonos-grouping', 'data', sGroupedData);
                storeSubParameter('sonos-grouping', 'sceneID', nSceneID);
              }
            }
          }
          else {
            lInfo.logln("Another sonos grouped scene is active, current scene "+parseInt(raisedEvent.parameter.sceneID)+" not executed");            
          }
          Property.store();
        }
      }
    }
    lInfo.logln("END sonos_execute");
  }
  catch (e) {
        lInfo.logln("Error by sonos_execute: " + e);
        for ( var i in e) {
            lInfo.logln(i + ': ' + e[i]);
        }                        
  }   
}                              
else if (raisedEvent.name === 'sonos_execute_undo') {
  try {    
    lInfo.logln("undoScene zoneID/groupID/sceneID:  " +  raisedEvent.source.zoneID +" / " +raisedEvent.source.groupID+ " / "+ raisedEvent.parameter.sceneID);

    var nGroupID =  parseInt("" + raisedEvent.source.groupID, 10);
    var nSceneID =  parseInt("" + raisedEvent.parameter.sceneID, 10);
    
    if (typeof raisedEvent.parameter.sceneID === 'undefined') {
       switch(raisedEvent.parameter.statename) {
         case 'fire':
           nSceneID = 76;
         break;
       }
    }
    
    var sMainPath = raisedEvent.parameter.path.replace("/undoCallScene","");
    lInfo.logln("Path "+sMainPath);
    var execAction = Property.getNode(sMainPath);

    if (execAction != null) {// && predefinedUnduEvents.indexOf(nSceneID) > -1)
      var restartData = null;
      
      // find sceneID in activities
      var arrayPlayers = getPlayersToExecUndo(execAction);
      var bGrouping = JSON.parse(execAction.getChild('createSonosGroup').getValue());
      
      if (bGrouping) {
        var baseNode = Property.getNode('sonos-grouping');
        if (baseNode != null) {
          arrayPlayers = null;
          if (baseNode.getChild('data') != null && baseNode.getChild('data').getValue() != '') {
            nStoredGroupSceneID = parseInt(baseNode.getChild('sceneID').getValue(),10); 
            if (nStoredGroupSceneID == nSceneID) {
              sValue = baseNode.getChild('data').getValue();
              if (sValue != "") {
                sValue = unescape(sValue);//replaceAll('&quot;','"',sValue); 
                arrayPlayers = JSON.parse(sValue);
              }
            }
          }
        }
      }                              
      lInfo.logln(">>>>>>>UNDO ExecPlayers:"+JSON.stringify(arrayPlayers));             
      
      if (arrayPlayers != null && arrayPlayers.length > 0) {       
        var arraySonosGroups = null;
        for (var iIndex=0;iIndex<arrayPlayers.length;iIndex++) {
          var sPlayerIP = arrayPlayers[iIndex].ip;
          var sDeviceID = arrayPlayers[iIndex].device;
          var sPlayerID = arrayPlayers[iIndex].id;
         
          if (bGrouping) { 
            if (arraySonosGroups == null)
              arraySonosGroups = upnpGetSonosGroups(sPlayerIP, curl_options);
          
            upnpSetPlayerAsStandAloneGroup(sPlayerIP, curl_options);
            addSlavePlayersToGroup(arrayPlayers[iIndex].slaves, arraySonosGroups, sDeviceID, curl_options);
          }
          
          var undoData = handleUndoActivity(sPlayerID,0,nSceneID);
          if (undoData != null) {
            var restartData = undoData.restoreData; 
        lInfo.logln("Undo restart data:"+JSON.stringify(restartData));             
            if (restartData != null) {
              if (restartData.activeQueueID != "") {
                // loadingOldQueue
                var aTmp = restartData.activeQueueID.split(':');
                if (aTmp != null && aTmp.length==2) {
                  var sQueueNr = aTmp[1];
                  upnpClearQueue(restartData.playerIP, curl_options);
                  upnpLoadPlaylist(restartData.playerIP, "file:///jffs/settings/savedqueues.rsq#"+sQueueNr, curl_options);                    
                }
                lInfo.logln("Deleting tmp queue");  
                upnpDeleteQueueFromList(restartData.playerIP, restartData.activeQueueID, curl_options);
              }
             
              if (restartData.volumeReset) {
                upnpSetVolume(restartData.playerIP, restartData.volumeDef, curl_options);
                
                if (arrayPlayers[iIndex].IPSlaves != null) {
                  lInfo.logln("Reset also slaves");
                  var aSlaves = null;
                  if (arrayPlayers[iIndex].IPSlaves.length > 0) {
                    aSlaves = arrayPlayers[iIndex].IPSlaves.split(","); 
                  }
                  
                  var aSlavesVolumes = null;
                  if (restartData.volumeSlaves != null && restartData.volumeSlaves.length > 0) {
                    aSlavesVolumes = restartData.volumeSlaves.split(","); 
                  }
                  lInfo.logln("Reset also slaves2");
                  if (aSlaves != null && aSlaves.length > 0) {
                    for (var iIndex=0;iIndex<aSlaves.length;iIndex++) {
                      var sSlaveIP = aSlaves[iIndex].trim();
                      
                      var nSlaveVolume = restartData.volumeDef;
                      if (aSlavesVolumes != null && aSlavesVolumes[iIndex] != null) {
                        var tmpVolume = parseInt(""+aSlavesVolumes[iIndex],10);
                        if (tmpVolume > -1) 
                          nSlaveVolume = tmpVolume;
                      }
                      
                      lInfo.logln("Reset slave volume for "+sSlaveIP+" to "+nSlaveVolume);
                      upnpSetVolume(sSlaveIP, nSlaveVolume, curl_options);
                    }
                  }
                }
              }
              upnpSetPlayMode(restartData.playerIP, restartData.playMode, curl_options);

              if (!restartData.isRadio) {
    lInfo.logln("Undo Restarting music:"+restartData.playerIP+"/x-rincon-queue:"+sDeviceID+"#0"+"/"+restartData.activeAction); 
               // activating queue
                upnpSetContentSrc(restartData.playerIP, "x-rincon-queue:"+sDeviceID+"#0", curl_options);
                upnpSeek(restartData.playerIP, true, restartData.trackListPos, curl_options);
                upnpSeek(restartData.playerIP, false, restartData.trackPlayPos, curl_options);
                if (restartData.activeAction == "PLAYING")
                  upnpPlay(restartData.playerIP, curl_options);
              } 
              else {
    lInfo.logln("Undo  Restarting radio:"+restartData.playUrl+"/"+restartData.activeAction); 
                upnpStop(restartData.playerIP, curl_options);
                upnpSetRadio(restartData.playerIP, restartData.playUrl, "Radio", curl_options);
                if (restartData.activeAction == "PLAYING")
                  upnpPlay(restartData.playerIP, curl_options);
              }
            }
            else {
              if (undoData.performStop) {
                lInfo.logln("No sonos player undo, stop event music");
                upnpStop(sPlayerIP, curl_options);
              }
            }
          }
        }
        
        if (bGrouping) {
          storeSubParameter('sonos-grouping', 'data', '');
          storeSubParameter('sonos-grouping', 'sceneID', '');
          Property.store();
        }
      }
    }

    lInfo.logln("end undo scene");
  }
  catch (e) {
        lInfo.logln("Error by undoScene: " + e);
        for ( var i in e) {
            lInfo.logln(i + ': ' + e[i]);
        }                        
  }    
  
}
else if (raisedEvent.name === 'sonos_execute_timer') {
     var timerData = JSON.parse(unescape(raisedEvent.parameter.path));
     //lInfo.logln(JSON.stringify(timerData));
     
     if (timerData != null) {
       // reseting dimming values
       resetDimming();
       
       var playerInfo = handleActivityTimeout(timerData.id, 0
                                              , timerData.sceneID, timerData.eventType);
       
       lInfo.logln("after exec handleActivityTimeout"); 
       if (playerInfo.restoreData != null) {
         if (timerData.onlyStop) {
           upnpStop(timerData.playerIP, curl_options);
         }
         else {
           lInfo.logln("handleActivityTimeout return data"); 
           var setSonosData = playerInfo.restoreData;
           if (setSonosData != null) {
           if (setSonosData.activeQueueID != "") {
             // loadingOldQueue
             var aTmp = setSonosData.activeQueueID.split(':');
             if (aTmp != null && aTmp.length==2) {
               var sQueueNr = aTmp[1];
               upnpClearQueue(setSonosData.playerIP, curl_options);
               upnpLoadPlaylist(setSonosData.playerIP, "file:///jffs/settings/savedqueues.rsq#"+sQueueNr, curl_options);                    
             }
             lInfo.logln("Deleting tmp queue");  
             upnpDeleteQueueFromList(setSonosData.playerIP, setSonosData.activeQueueID, curl_options);
           }
          
           lInfo.logln("Restart2");                      
           if (setSonosData.volumeReset) {
             lInfo.logln("Reset volume to "+setSonosData.volumeDef);                      
             upnpSetVolume(setSonosData.playerIP, setSonosData.volumeDef, curl_options);
             //lInfo.logln("Reseted volume "+upnpGetVolume(setSonosData.playerIP, curl_options));
           }
           if (!setSonosData.isRadio) {
             lInfo.logln("Restarting music:"+"x-rincon-queue:"+timerData.deviceID+"#0"+"/"+setSonosData.activeAction); 
             // activating queue
             upnpSetContentSrc(setSonosData.playerIP, "x-rincon-queue:"+timerData.deviceID+"#0", curl_options);
             upnpSeek(setSonosData.playerIP, true, setSonosData.trackListPos, curl_options);
             upnpSeek(setSonosData.playerIP, false, setSonosData.trackPlayPos, curl_options);
             if (setSonosData.activeAction == "PLAYING")
               upnpPlay(setSonosData.playerIP, curl_options);
           } 
           else {
             lInfo.logln("Restarting radio:"+setSonosData.playUrl+"/"+setSonosData.activeAction); 
             upnpStop(setSonosData.playerIP, curl_options);
             upnpSetRadio(setSonosData.playerIP, setSonosData.playUrl, "Radio", curl_options);
             if (setSonosData.activeAction == "PLAYING")
               upnpPlay(setSonosData.playerIP, curl_options);
           }
         }
         }          
       }
       else if (playerInfo.activityId != -1) {
         lInfo.logln("handleActivityTimeout return activityID="+playerInfo.activityId)
         var execAction = Property.getNode('/scripts/sonos-controller/sonos-activities/'+playerInfo.activityId);
         executeComandOnSonos(timerData.playerIP, timerData.deviceID, execAction, curl_options, "", 'sonos-activities');
         upnpSetPlayMode(timerData.playerIP, "REPEAT_ALL", curl_options);
       }     
       else if (playerInfo.performStop) {
         lInfo.logln("No data, stop event music");
         upnpStop(timerData.playerIP, curl_options);
       }
       
       if (timerData.sceneID == -1) { 
         var evt = new Event("sonos-test-activity-finished",  {});
         evt.raise(); 
       }
     }
  }
else if (raisedEvent.name === 'sonos_execute_timer_group') {
      var timerData = JSON.parse(unescape(raisedEvent.parameter.path));
      lInfo.logln("Grouped speakers timer finishing="+JSON.stringify(timerData));
       
      if (timerData != null) { 
        // reseting dimming values
        resetDimming();     
        
        // test if scene not all-ready finished  
        var bSceneStillActive = false;  
        var baseNode = Property.getNode('sonos-grouping');
        if (baseNode != null) {
          arrayPlayers = null;
          if (baseNode.getChild('data') != null && baseNode.getChild('data').getValue() != '') {
            nStoredGroupSceneID = parseInt(baseNode.getChild('sceneID').getValue(),10); 
            if (nStoredGroupSceneID == timerData.sceneID) {
              sValue = baseNode.getChild('data').getValue();
              if (sValue != "")
                bSceneStillActive = true;
            }
          }
        }
       
        if (bSceneStillActive && timerData.playerMasters != null) {
          lInfo.logln("For sceneID="+timerData.sceneID+", players count="+timerData.playerMasters.length);
          var arraySonosGroups = null;
          for (var iIndex=0;iIndex<timerData.playerMasters.length;iIndex++) {
            var player = timerData.playerMasters[iIndex];
            var playerInfo = handleActivityTimeout(player.id, 0, timerData.sceneID, timerData.eventType);
   
            lInfo.logln("after exec handleActivityTimeout, restoreData not null");          
            if (iIndex==0) 
              arraySonosGroups = upnpGetSonosGroups(player.ip, curl_options);
   
            // restoring sonos groups
            // set player as standalone in group
            lInfo.logln("Set as standalone "+player.ip);    
            upnpSetPlayerAsStandAloneGroup(player.ip, curl_options);
            // check if has sub players, add them also to group
            addSlavePlayersToGroup(player.slaves, arraySonosGroups, player.device, curl_options);

            if (playerInfo.restoreData != null) {
                if (timerData.onlyStop) {
                  upnpStop(player.ip, curl_options);
                }
                else {
                  lInfo.logln("handleActivityTimeout return data"); 
                  var setSonosData = playerInfo.restoreData;
                  if (setSonosData != null) {
                    if (setSonosData.activeQueueID != "") {
                      // loadingOldQueue
                      var aTmp = setSonosData.activeQueueID.split(':');
                      if (aTmp != null && aTmp.length==2) {
                        var sQueueNr = aTmp[1];
                        upnpClearQueue(setSonosData.playerIP, curl_options);
                        upnpLoadPlaylist(setSonosData.playerIP, "file:///jffs/settings/savedqueues.rsq#"+sQueueNr, curl_options);                    
                      }
                      lInfo.logln("Deleting tmp queue");  
                      upnpDeleteQueueFromList(setSonosData.playerIP, setSonosData.activeQueueID, curl_options);
                    }
                   
                    lInfo.logln("Restart2");                      
                    if (setSonosData.volumeReset) {
                      lInfo.logln("Reset volume to "+setSonosData.volumeDef);                      
                      upnpSetVolume(setSonosData.playerIP, setSonosData.volumeDef, curl_options);
                      lInfo.logln("New volume "+upnpGetVolume(setSonosData.playerIP, curl_options));
                      
                      // set volume also for all slave players in group
                      if (player.IPSlaves != null) {
                        lInfo.logln("Reset also slaves");
                        var aSlaves = null;
                        if (player.IPSlaves.length > 0) {
                          aSlaves = player.IPSlaves.split(","); 
                        }
                        
                        var aSlavesVolumes = null;
                        if (setSonosData.volumeSlaves != null && setSonosData.volumeSlaves.length > 0) {
                          aSlavesVolumes = setSonosData.volumeSlaves.split(","); 
                        }
                        lInfo.logln("Reset also slaves2");
                        if (aSlaves != null && aSlaves.length > 0) {
                          for (var iIndex=0;iIndex<aSlaves.length;iIndex++) {
                            var sSlaveIP = aSlaves[iIndex].trim();
                            
                            var nSlaveVolume = setSonosData.volumeDef;
                            if (aSlavesVolumes != null && aSlavesVolumes[iIndex] != null) {
                              var tmpVolume = parseInt(""+aSlavesVolumes[iIndex],10);
                              if (tmpVolume > -1) 
                                nSlaveVolume = tmpVolume;
                            }
                            
                            lInfo.logln("Reset slave volume for "+sSlaveIP+" to "+nSlaveVolume);
                            upnpSetVolume(sSlaveIP, nSlaveVolume, curl_options);
                          }
                        }
                      }
                    }
                    if (!setSonosData.isRadio) {
                      lInfo.logln("Restarting music:"+"x-rincon-queue:"+player.device+"#0"+"/"+setSonosData.activeAction); 
                      // activating queue
                      upnpSetContentSrc(setSonosData.playerIP, "x-rincon-queue:"+player.device+"#0", curl_options);
                      upnpSeek(setSonosData.playerIP, true, setSonosData.trackListPos, curl_options);
                      upnpSeek(setSonosData.playerIP, false, setSonosData.trackPlayPos, curl_options);
                      if (setSonosData.activeAction == "PLAYING")
                        upnpPlay(setSonosData.playerIP, curl_options);
                    } 
                    else {
                      lInfo.logln("Restarting radio:"+setSonosData.playUrl+"/"+setSonosData.activeAction); 
                      upnpStop(setSonosData.playerIP, curl_options);
                      upnpSetRadio(setSonosData.playerIP, setSonosData.playUrl, "Radio", curl_options);
                      if (setSonosData.activeAction == "PLAYING")
                        upnpPlay(setSonosData.playerIP, curl_options);
                    }
                  }
                }
              }
              else if (playerInfo.activityId != -1) {
                lInfo.logln("handleActivityTimeout return activityID="+playerInfo.activityId)
                var execAction = Property.getNode('/scripts/sonos-controller/sonos-activities/'+playerInfo.activityId);
                executeComandOnSonos(player.ip, player.device, execAction, curl_options, "", 'sonos-activities');
                upnpSetPlayMode(player.ip, "REPEAT_ALL", curl_options);
             }  
           }
           storeSubParameter('sonos-grouping', 'data', '');
           storeSubParameter('sonos-grouping', 'sceneID', '');
           Property.store();
         }
         
        if (timerData.sceneID == -1) { 
          var evt = new Event("sonos-test-activity-finished",  {});
          evt.raise(); 
        }
      }
    }
